(* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *  Copyright (C) 2004-2006 Ekrem "ECoder" KOAK                        *
 *  Mail: ekremkocak@hotmail.com                                                 *
 *                                                                           *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *)


unit PushSource;

interface
 uses
   BaseClass, DirectShow9, DSUtil, Windows, ActiveX, SysUtils,ExtCtrls,
    Graphics, Registry,Dialogs,Classes;

  const
   IID_ITSPushSource: TGUID ='{92F75CE4-D9D2-455A-8F9A-BD33D52E4586}';
   CLSID_TSPushSource: TGuid = '{173C2533-014E-42C2-93EE-CB6EA5914FFE}';

   TS_PACKET_SIZE  =188;
   MAX_BUFFER_SIZE =188 * 512;

  type

   ITSPushSource = interface(IUnknown)
  ['{92F75CE4-D9D2-455A-8F9A-BD33D52E4586}']
    function PushData(ABuffer: PByte; ASize: Integer): HRESULT; stdcall;
  end;





  TSPushSourceOutPin= class(TBCSourceStream)
  protected
   public
     constructor Create(const ObjectName: AnsiString; out hr: HRESULT; Filter: TBCSource; const Name: WideString);
     destructor Destroy; override;
       function CheckMediaType(MediaType: PAMMediaType): HRESULT; override;
       function GetMediaType(MediaType: PAMMediaType): HResult; override;
       function DecideBufferSize(Allocator: IMemAllocator; Properties: PAllocatorProperties): HRESULT; override;
       function  FillBuffer(Sample: IMediaSample): HRESULT; override;
       procedure ReceiveTS(ABuffer: PByte; ASize: Integer);
  end;

  TSPushSource = class (TBCSource,ITSPushSource,IPersist)
  protected
    FPin: TSPushSourceOutPin;
  public
     constructor Create();
     destructor Destroy; override;
     function PushData(ABuffer: PByte; ASize: Integer): HRESULT; stdcall;

  end;

implementation

constructor TSPushSourceOutPin.Create(const ObjectName: AnsiString; out hr: HRESULT; Filter: TBCSource; const Name: WideString);
begin
 inherited Create('Out', hr, Filter,'Out');
end;


destructor TSPushSourceOutPin.Destroy;
begin
  inherited Destroy;
end;

function TSPushSourceOutPin.DecideBufferSize(Allocator: IMemAllocator; Properties: PAllocatorProperties): HRESULT;
var
  Actual: ALLOCATOR_PROPERTIES;
begin
   FFilter.StateLock.Lock;
 try

   Result := E_FAIL;
  if (Allocator = nil) or (Properties = nil) then
  begin
    Result := E_POINTER;
    Exit;
  end;
 if (Properties.cBuffers = 0) then
      Properties.cBuffers := 1;
      Properties.cbBuffer :=MAX_BUFFER_SIZE;
      Properties.cbAlign :=TS_PACKET_SIZE;
      Properties.cbPrefix := 0;

    Result := Allocator.SetProperties(Properties^, Actual);
    if Failed(Result)
      then Exit;

    if (Actual.cbBuffer < Properties.cbBuffer)
      then Result := E_FAIL
      else Result := S_OK;

  finally
   FFilter.StateLock.Unlock;
 end;

end;

function TSPushSourceOutPin.FillBuffer(Sample: IMediaSample): HRESULT;
begin
  Result := S_FALSE;
end;

function TSPushSourceOutPin.GetMediaType(MediaType: PAMMediaType): HResult;
begin
   FFilter.StateLock.Lock;
  try
     MediaType.MajorType            := MEDIATYPE_Stream;
     MediaType.SubType              := MEDIASUBTYPE_MPEG2_TRANSPORT;
     MediaType.bFixedSizeSamples    := True;
     MediaType.formattype           := FORMAT_None;
     MediaType.bTemporalCompression := False;
     MediaType.cbFormat             := 0;
     MediaType.pbFormat             := nil;

   Result := S_OK;
  finally
    FFilter.StateLock.Unlock;
  end;
end;

function TSPushSourceOutPin.CheckMediaType(MediaType: PAMMediaType): HRESULT;
begin
  Result := S_OK;
end;

procedure TSPushSourceOutPin.ReceiveTS(ABuffer: PByte; ASize: Integer);
var
  sample: IMediaSample;
  buffer: PByte;
  hr: HRESULT;
  size: Integer;
  max_size: Integer;
begin
  if not IsConnected then
  begin
    Exit;
  end;

  if not (FFilter.State = State_Running) then
  begin
    Exit;
  end;

  while ASize > 0 do
  begin
    hr := Self.GetDeliveryBuffer(sample, nil, nil, 0);

    if (hr <> S_OK) then
    begin
      Exit;
    end;

    sample.SetDiscontinuity(False);
    hr := sample.GetPointer(buffer);
    if (hr <> S_OK) then
    begin
      Exit;
    end;

    max_size := sample.GetSize;
    if max_size <= 0 then
    begin
      Exit;
    end;

    if (ASize > max_size)
      then size := max_size
      else size := ASize;
      
    sample.SetActualDataLength(size);
    Move(ABuffer^, buffer^, size);
    hr := Deliver(sample);

    if hr <> S_OK then
    begin
      Exit;
    end;

    inc(ABuffer, size);
    dec(ASize, size);
  end;
end;

(*** TSPushSource **************************************************)

constructor TSPushSource.Create();
var
 Unk: IUnKnown;
  hr: HRESULT ;
begin
  inherited Create('TSPushSource', unk, CLSID_TSPushSource, hr);
   FPin := TSPushSourceOutPin.Create('Output', hr, Self, 'Output');
   if not Assigned(FPin) then hr := S_FALSE
end;


destructor TSPushSource.Destroy;
begin
  FPin.Free;
  inherited Destroy;
end;

function TSPushSource.PushData(ABuffer: PByte; ASize: Integer): HRESULT;
begin
 Result    := S_FALSE;
  if FPin<>nil then
     FPin.ReceiveTS(ABuffer, ASize);
  Result := S_OK;

end;


end.


