unit UDVBScan;

interface
uses SysUtils, Windows, Classes;

const
  BUF_SIZE = 4096;
  PACKET_SIZE =188;
  PACKET_NUM =1;

  MAX_PACKET_SIZE = 4096;
  MAX_PAT_RECORDS = 100;
  MAX_TELETEXT_RECORDS = 100;
  MAX_AUDIO_PIDS = 100;
  MAX_CA_PIDS = 100;

  PAT_WAIT_TIME = 10000;
  PMT_WAIT_TIME = 10000;
  SDT_WAIT_TIME = 30000;


 Const
  DISEQC_HIGH_NIBLE  = $F0;
  DISEQC_LOW_BAND    = $00;
  DISEQC_HIGH_BAND   = $01;
  DISEQC_VERTICAL    = $00;
  DISEQC_HORIZONTAL  = $02;
  DISEQC_POSITION_A  = $00;
  DISEQC_POSITION_B  = $04;
  DISEQC_OPTION_A    = $00;
  DISEQC_OPTION_B    = $08;


Const
  TP_NAME_LEN = 64;
  TP_AUDIO_COUNT = 16;
  TP_CHANNAL_COUNT = 128;
  TP_CA_COUNT = 64;


Type
  TCAInfo = packed record
    wCount:    WORD;
    wCAType:   packed array [0..TP_CA_COUNT-1] of WORD;
    wPID:      packed array [0..TP_CA_COUNT-1] of WORD;
  end;
Type
  TAudioPID = packed record
    wCount:     WORD;
    wPID:       packed array [0..TP_AUDIO_COUNT-1] of WORD;
    Language:   packed array [0..TP_AUDIO_COUNT-1, 0..2] of char;
  end;

Type
  TChannelInfo = packed record
    wSID:           WORD;       //Channel ID
    wPMT:           WORD;       //PID for Program Map Table
    wType:          WORD;
    wScrambled:     WORD;
    ChannelName:    packed array [0..TP_NAME_LEN-1] of char;
    ProviderName:   packed array [0..TP_NAME_LEN-1] of char;
    wPCR:           WORD;       //PID for Program Clock Reference
    wVideo:         WORD;       //Video PID
    Audio:          TAudioPID;  //Array of Audio PIDs
    wAC3PID:        WORD;
    wTeletext:      WORD;
    ECM:            TCAInfo;
  end;

  TTransporderInfo = packed record
    ChannelCount: Integer;
    Channel: packed array[0..TP_CHANNAL_COUNT-1] of TChannelInfo;
    TSID: Word;         //Transport Stream ID
    wNetID: Word;       //Network ID
  end;
  pTransporderInfo = ^TTransporderInfo;


type
  TDVBSection=class(TObject)
    ScannedPID: Integer;
    StopTime: TDateTime;
    RecordCount: Integer;
    FilterHandle: DWord;
    Loop: Integer;
    Active: Boolean;
    Completed: Boolean;
    Offset: Integer;
    SectionSize: Integer;
    Packet: array [0..MAX_PACKET_SIZE-1] of Byte;
    destructor Destroy; override;
    function IsValidSectionStart(Buff: PByteArray): Boolean;virtual;
    function ProcessSection: Boolean;virtual;
    procedure Packetize(Data: PByteArray);
    procedure ProcessPackets(Length: Cardinal; Data: PByteArray);
    function Start(PID: Integer): Boolean;
    procedure StopScan;
  end;

  TPATSection=class(TDVBSection)
    NetworkPID: Integer;
    TSID: Integer;
    SIDs: array [1..MAX_PAT_RECORDS] of integer;
    PMTs: array [1..MAX_PAT_RECORDS] of integer;
    function IsValidSectionStart(Buff: PByteArray): Boolean;override;
    function ProcessSection: Boolean;override;
    procedure InitSection;
  end;

  TPMTSection=class(TDVBSection)
    ScannedSID : Integer;
    VideoPID : Integer;
    TeletextPID: Integer;
    PCRPid: Integer;
    AudioPIDCount: Integer;
    AudioPIDs: array[1..MAX_AUDIO_PIDS] of Word;
    AudioLanguages: array[1..MAX_AUDIO_PIDS, 1..3] of char;
    AC3Pid: Integer;
    CACount: Integer;
    CAIDs: array[1..MAX_CA_PIDS] of Integer;
    CAECMs: array[1..MAX_CA_PIDS] of Integer;
    TeletextRecordCount: Integer;
    TeletextLanguages: array[1..MAX_TELETEXT_RECORDS, 1..3] of char;
    TeletextTypes: array[1..MAX_TELETEXT_RECORDS] of Integer;
    TeletextPages: array[1..MAX_TELETEXT_RECORDS] of Integer;
    function IsValidSectionStart(Buff: PByteArray): Boolean;override;
    function ProcessSection: Boolean;override;
    procedure InitSection;
  end;

  TSDTSection=class(TDVBSection)
    SearchedTSID: Integer;
    NetworkID: Integer;
    SIDs: array [1..MAX_PAT_RECORDS] of Integer;
    CAmode: array [1..MAX_PAT_RECORDS] of Byte;
    Providers: array [1..MAX_PAT_RECORDS] of String;
    Services: array [1..MAX_PAT_RECORDS] of String;
    function IsValidSectionStart(Buff: PByteArray): Boolean;override;
    function ProcessSection: Boolean;override;
    procedure InitSection;
  end;

function DVBGetTransporderInfo(TPInfo: PTransporderInfo): Integer;


var
  DVBSectionList: TList;
  LockDVBSectionList: TRTLCriticalSection;
  PATSection: TPATSection;
  PMTSection: TPMTSection;
  PMTSectionList: TList;
  SDTSection: TSDTSection;
  ScanStop: Boolean;

  TransporderInfo: TTransporderInfo;
  Buf: array[0..BUF_SIZE-1] of char;
  BufCount:integer;
implementation
uses Math, Forms,SkyStar2;

const
  PID_PAT = $00;
  TID_PAT = $00;
  TID_PMT = $02;
  PID_SDT = $11;
  TID_SDT_ACT = $42;
  TID_SDT_OTH = $46;

type
  TDescriptor=packed record
    Tag: Byte;
    Len: Byte;
    Data: array[0..100] of Byte;
  end;
  PDescriptor = ^TDescriptor;


procedure DVBSectionCallBack( Handle: DWORD; Length: Cardinal;Data: PByteArray);  stdcall;
var
  I: Integer;
  DVBSection: TDVBSection;
   header_size:integer;
  stream_id: byte;
begin
  EnterCriticalSection(LockDVBSectionList);
  try
    for I := 0 to DVBSectionList.Count - 1 do
    begin
      DVBSection := DVBSectionList[I];
      if DVBSection <> nil then
        with DVBSection do
          if Active and (FilterHandle = Handle) then
          begin
            //-----------------------------------------------------------
             header_size:=4;
             if (Data[3] and $20)>0 then  header_size:= header_size + (Data[4]+1);
               if (Data[1] and $40)>0 then
               begin
                 if (Data[header_size]=$00)and(Data[header_size]=$00)and(Data[header_size]=$01)then
                 begin
                 stream_id:=Data[header_size+3];
                   if (stream_id<>$BC)and(stream_id<>$BE)and(stream_id<>$BF)and(stream_id<>$F0)and(stream_id<>$F1)and(stream_id<>$F2)and(stream_id<>$F8)
                   then
                   header_size:= header_size + (8+Data[header_size+8]+1)
                   else
                  header_size:= header_size + (6);
                end;
             end;
         copymemory(@Buf[BufCount],@Data[header_size],PACKET_SIZE-header_size);
        BufCount:=BufCount + (PACKET_SIZE-header_size);
        if BufCount>=PACKET_SIZE*PACKET_NUM-188
       then
       begin
       ProcessPackets(BufCount,@Buf[0]);
       BufCount:= 0;
       exit;
      end;
     end;
    end;
  finally
    LeaveCriticalSection(LockDVBSectionList);
  end;
end;

destructor TDVBSection.Destroy;
begin
  StopScan;

  inherited Destroy;
end;




function TDVBSection.Start(PID: Integer): Boolean;
begin
  Result := False;
  ScannedPID := PID;
  Completed := False;
  RecordCount := 0;
  Loop := 0;
  Offset := 0;
  FilterHandle := SetFilter(PID,@DVBSectionCallBack);
  
  if FilterHandle = 0 then Exit;
  Active := True;
  EnterCriticalSection(LockDVBSectionList);
  try
    DVBSectionList.Add(Self);
  finally
    LeaveCriticalSection(LockDVBSectionList);
  end;
  Result := True;
end;

procedure TDVBSection.StopScan;
begin
  if Active then
  begin

 DelFilter(FilterHandle);

    Active := False;
    EnterCriticalSection(LockDVBSectionList);
    try
      DVBSectionList.Remove(Self);
    finally
      LeaveCriticalSection(LockDVBSectionList);
    end;
  end;
end;

function TDVBSection.IsValidSectionStart(Buff: PByteArray): Boolean;
begin
  Result := True;
end;

function TDVBSection.ProcessSection: Boolean;
begin
  Result := True;
end;

procedure TDVBSection.Packetize(Data: PByteArray);
var
  I: Integer;
  Flag: Boolean;
begin
    if not Active then Exit;
    if Offset = 0 then
    begin
      if not IsValidSectionStart(Data) then Exit;
      SectionSize := (Data[2] and $0F) * $100 + Data[3];
      if SectionSize = 0 then Exit;
    end;
    for I := 0 to Min(184, SectionSize - Offset + 4) - 1 do
      Packet[Offset + I] := Data[I];
    Offset := Offset + Min(184, SectionSize - Offset + 4);
    if Offset >= SectionSize + 4 then
    begin
      if Loop = Packet[7] then
      begin
        Flag := ProcessSection;
        if Flag then
        begin
          if Loop >= Packet[8] then
          begin
            StopScan;
            Completed := True;
          end
          else
            Loop := Loop + 1;
        end;
      end;
      Offset := 0;
    end;
end;

procedure TDVBSection.ProcessPackets(Length: Cardinal; Data: PByteArray);
var
  I: Cardinal;
begin
  I := 0;
  while (I < Length) and Active do
  begin
    Packetize(@Data[I]);
    I := I + 184;
  end;
end;

procedure TPATSection.InitSection;
var
  I: Integer;
begin
  ScannedPID := 0;
  RecordCount := 0;
  FilterHandle := 0;
  Loop := 0;
  Active := FALSE;
  Completed := FALSE;
  Offset := 0;
  SectionSize := 0;
  NetworkPID := 0;
  TSID := 0;
  for I := 1 to MAX_PAT_RECORDS do SIDs[I] := 0;
  for I := 1 to MAX_PAT_RECORDS do PMTs[I] := 0;
end;

function TPATSection.IsValidSectionStart(Buff: PByteArray): Boolean;
begin
  Result := (Buff[0] = 0) and (Buff[1] = TID_PAT);
end;

function TPATSection.ProcessSection: Boolean;
type
  TPATProg=packed record
    ProgramNumberHi: Byte;
    ProgramNumberLo: Byte;
    BitField1: Byte;
    NetworkPIDLo: Byte;
  end;

  TPAT=packed record
    TableID: Byte;
    BitField1:byte;
    SectionLengthLo: Byte;
    TransportStreamIDHi: Byte;
    TransportStreamIDLo: Byte;
    BitField2: Byte;
    SectionNumber: Byte;
    LastSectionNumber: Byte;
  end;
  PPAT = ^TPAT;
  PPATProg = ^TPATProg;

var
  PAT: PPAT;
  I, NumProgs, SectionLen: Integer;
  PATProg: PPATProg;
  SID,PMT: Integer;
begin
  Result := True;
  PAT := @Packet[1];
  TSID := PAT^.TransportStreamIDHi * $100 + PAT^.TransportStreamIDLo;
  SectionLen := ((PAT^.BitField1 and $0F) * $100) + PAT^.SectionLengthLo;
  NumProgs := (SectionLen - 8 ) div 4;
  for I := 0 to NumProgs - 1 do
  begin
    PATProg := @Packet[8 + I * 4 + 1];
    SID := PATProg^.ProgramNumberHi * $100 + PATProg^.ProgramNumberLo;
    PMT := (PATProg^.BitField1 and $1F) * $100 + PATProg^.NetworkPIDLo;
    if SID = 0 then
      NetworkPID := PMT
    else
    begin
      RecordCount := RecordCount + 1;
      SIDs[RecordCount] := SID;
      PMTs[RecordCount] := PMT;
    end;
  end;
end;

procedure TPMTSection.InitSection;
var
  I: integer;
begin
  ScannedPID := 0;
  RecordCount := 0;
  FilterHandle := 0;
  Loop := 0;
  Active := FALSE;
  Completed := FALSE;
  Offset := 0;
  SectionSize := 0;
  ScannedSID := 0;
  VideoPID := 0;
  TeletextPID := 0;
  PCRPid := 0;
  AudioPIDCount := 0;
  AC3Pid := 0;
  CACount := 0;
  TeletextRecordCount := 0;
  for I := 1 to MAX_AUDIO_PIDS do
  begin
    AudioPIDs[I] := 0;
    AudioLanguages[I,1]:=' ';
    AudioLanguages[I,2]:=' ';
    AudioLanguages[I,3]:=' ';
  end;
  for I := 1 to MAX_CA_PIDS do
  begin
    CAIDs[I] := 0;
    CAECMs[I] := 0;
  end;
  for I := 1 to MAX_TELETEXT_RECORDS do
  begin
    TeletextLanguages[I,1] := ' ';
    TeletextLanguages[I,2] := ' ';
    TeletextLanguages[I,3] := ' ';
    TeletextTypes[I] := 0;
    TeletextPages[I] := 0;
  end;
end;

function TPMTSection.IsValidSectionStart(Buff: PByteArray): Boolean;
begin
  Result := (Buff[0] = 0) and (Buff[1] = TID_PMT);
end;

function TPMTSection.ProcessSection: Boolean;
const
  PMT_LEN = 12;
type
  TPMTInfo=packed record
    StreamType: Byte;
    BitField1: Byte;
    ElementaryPIDLo: Byte;
    BitField2: Byte;
    ESInfoLengthLo: Byte;
  end;
  PPMTInfo = ^TPMTInfo;
  TPMT=packed record
    TableID: Byte;
    BitField1: Byte;
    SectionLengthLo: Byte;
    ProgramNumberHi: Byte;
    ProgramNumberLo: Byte;
    BitField2: Byte;
    SectionNumber: Byte;
    LastSectionNumber: Byte;
    BitField3: Byte;
    PCRPIDLo: Byte;
    BitField4: Byte;
    ProgramInfoLengthLo: Byte;
  end;
  PPMT = ^TPMT;

  procedure ParsePMTDescriptors(StreamType: Integer; PID: Integer; Ptr: PByteArray; Len: Integer);
  var
    Descriptor: PDescriptor;
    CurrentPos: Integer;
    I: Integer;
  begin
    CurrentPos := 0;
    while CurrentPos < Len do
    begin
      Descriptor := @Ptr[CurrentPos];
      if (StreamType = 6) and (Descriptor.Tag = $56) then
      begin
        TeletextPID := PID;
        TeletextRecordCount := Descriptor.Len div 5;
        for I := 1 to TeletextRecordCount do
        begin
          TeletextLanguages[I,1] := Chr(Descriptor.Data[I * 5 - 5 + 0]);
          TeletextLanguages[I,2] := Chr(Descriptor.Data[I * 5 - 5 + 1]);
          TeletextLanguages[I,3] := Chr(Descriptor.data[I * 5 - 5 + 2]);
          TeletextTypes[I] := Descriptor.Data[I * 5 - 5 + 3] div 8;
          TeletextPages[I] := (Descriptor.Data[I * 5 - 5 + 3] and 7) * 100
            +((Descriptor.Data[I * 5 - 5 + 4] and $F0) div $10) * 10
            +(Descriptor.Data[I * 5 - 5 + 4] and $0F);
        end;
      end
      else if (StreamType in [3,4]) and (Descriptor.Tag = $0A) then
      begin
        AudioLanguages[AudioPIDCount,1] := Chr(Descriptor.Data[0]);
        AudioLanguages[AudioPIDCount,2] := Chr(Descriptor.Data[1]);
        AudioLanguages[AudioPIDCount,3] := Chr(Descriptor.Data[2]);
      end
      else if (Descriptor.Tag = $09) then
      begin
        CACount := CACount + 1;
        CAIDs[CACount] := Descriptor.Data[0] * $100 + Descriptor.Data[1];
        CAECMs[CACount] := (Descriptor.Data[2] and $1F) * $100 + Descriptor.Data[3];
      end;
      if (StreamType = 6) and (Descriptor.Tag = $6A) then
      begin
        AC3Pid := PID;
      end;
      CurrentPos := CurrentPos + 2 + Descriptor.Len;
    end;
  end;

var
  PMT: PPMT;
  PMTInfo: PPMTInfo;
  Ptr: PByteArray;
  StreamLen, InfoLength, ProgNum, StreamType, PID: Integer;
  ProgramInfoLength: Integer;
begin
  VideoPID := 0;
  TeletextPID := 0;
  AudioPIDCount := 0;
  TeletextRecordCount := 0;
  AC3Pid := 0;
  Result := True;
  PMT := @Packet[1];
  CACount := 0;
  ProgramInfoLength := (PMT.BitField4 and $0F) * $100 + PMT.ProgramInfoLengthLo;
  ProgNum := PMT.ProgramNumberHi * $100 + PMT.ProgramNumberLo;
  PCRPid := (PMT.BitField3 and $1F) * $100 + PMT.PCRPIDLo;
  Ptr := @Packet[1 + PMT_LEN];
  ParsePMTDescriptors(-1, 0, Ptr, ProgramInfoLength);
  Ptr := @Ptr[ProgramInfoLength];
  StreamLen := (SectionSize - ProgramInfoLength - PMT_LEN - 4);
  if ProgNum <> ScannedSID then
    Result := False
  else
  begin
    while (StreamLen > 0) do
    begin
      PMTInfo := @Ptr[0];
      Ptr := @Ptr[5];
      StreamType := PMTInfo.StreamType;
      PID := ((PMTInfo.BitField1 and $1F) * $100) + PMTInfo.ElementaryPIDLo;
      InfoLength := ((PMTInfo.BitField2 and $0F) * $100) + PMTInfo.ESInfoLengthLo;
      if (StreamType = 2) then VideoPID := PID
      else if StreamType in [3,4] then
      begin
        AudioPIDCount := AudioPIDCount + 1;
        AudioPIDs[AudioPIDCount] := PID;
        AudioLanguages[AudioPIDCount] := '   ';
      end;
      ParsePMTDescriptors(StreamType, PID, Ptr, InfoLength);
      StreamLen := StreamLen - 5 - InfoLength;
      if InfoLength > 0 then Ptr := @Ptr[InfoLength];
    end;
  end;
end;

procedure TSDTSection.InitSection;
var
  I: Integer;
begin
  ScannedPID := 0;
  RecordCount := 0;
  FilterHandle := 0;
  Loop := 0;
  Active := False;
  Completed := False;
  Offset := 0;
  SectionSize := 0;
  SearchedTSID := 0;
  NetworkID := 0;
  for I := 1 to MAX_PAT_RECORDS do
  begin
    SIDs[I] := 0;
    Providers[I] := '';
    Services[I] := '';
  end;
end;

function TSDTSection.IsValidSectionStart(Buff: PByteArray): Boolean;
begin
  Result := (Buff[0] = 0) and (Buff[1] = TID_SDT_ACT) and (SearchedTSID = Buff[4] * $100 + Buff[5]);
end;

procedure iso2win(var Str: String);
const
  Chars: array[#161..#241] of Char = (#168, #162, #163, #164, #165, #166, #167, #168, #169, #170, #171, #172, #173, #174, #175, #192, #193, #194, #195, #196, #197, #198, #199, #200, #201, #202, #203, #204, #205, #206, #207, #208, #209, #210, #211, #212, #213, #214, #215, #216, #217, #218, #219, #220, #221, #222, #223, #224, #225, #226, #227, #228, #229, #230, #231, #232, #233, #234, #235, #236, #237, #238, #239, #240, #241, #242, #243, #244, #245, #246, #247, #248, #249, #250, #251, #252, #253, #254, #255, #240, #184);
var
  I: Integer;
begin
  for I := 1 to Length(Str) do
    if (Str[I] >= #161) and (Str[I] <= #241) then
      Str[I] := Chars[Str[I]];
end;

function TSDTSection.ProcessSection: Boolean;
const
  SDT_LEN = 11;
  SDT_DESCR_LEN = 5;
  DESCR_GEN_LEN = 2;
type
  TSDTDescriptor=packed record
    ServiceIDHi: Byte;
    ServiceIDLo: Byte;
    BitField1: Byte;
    BitField2: Byte;
    DescriptorsLoopLengthLo: Byte;
  end;
  PSDT = ^TSDT;

  TSDT=packed record
    TableID: Byte;
    BitField1: Byte;
    SectionLengthLo: Byte;
    TransportStreamIDHi: Byte;
    TransportStreamIDLo:byte;
    BitField2: Byte;
    SectionNumber: Byte;
    LastSectionNumber: Byte;
    OriginalNetworkIDHi: Byte;
    OriginalNetworkIDLo: Byte;
    Buffer: Byte;
  end;
  PSDTDescriptor =^ TSDTDescriptor;

var
  SDT: PSDT;
  Ptr, Ptr1: PByteArray;
  SDTDescriptor: PSDTDescriptor;
  SecLen: Integer;
  SIDID: Integer;
  CAm: Byte;
//  ServiceType: Integer;
  Len, LoopLen, DescrLen: Integer;
  Str: String;
begin
  Result := True;
  SDT := @Packet[1];
  Ptr := @Packet[1];
  Len := 0;
  SecLen := ((SDT^.BitField1 and $0F) * $100) + SDT^.SectionLengthLo;
  if not (SDT^.TableID in [TID_SDT_ACT,TID_SDT_OTH]) then exit;
  NetworkID := SDT^.OriginalNetworkIDHi * $100 + SDT^.OriginalNetworkIDLo;
  Ptr := @Ptr[SDT_LEN];
  Len := Len + SDT_LEN;
  while Len < SecLen - 1 do
  begin
    SDTDescriptor := @Ptr[0];
    SIDID := SDTDescriptor^.ServiceIDHi * $100 + SDTDescriptor^.ServiceIDLo;
    CAm := (SDTDescriptor^.BitField2 and $10) div 16;
    LoopLen := ((SDTDescriptor^.BitField2 and $0F) * $100 ) + SDTDescriptor^.DescriptorsLoopLengthLo;
    Ptr := @Ptr[SDT_DESCR_LEN];
    Len := Len + SDT_DESCR_LEN + LoopLen;
    DescrLen := 0;
    while (DescrLen < LoopLen) do
    begin
      case Ptr[0] {Descriptor} of
        $48: //Service description
          begin
//            ServiceType := Ptr[2];
            RecordCount := RecordCount + 1;
            SIDs[RecordCount] := SIDID;
            CAmode[RecordCount] := CAm;
            if Ptr[3] > 0 then
            begin
              if Ptr[4] >= $20 then
                Providers[RecordCount] := copy(PChar(@Ptr[4]), 1, Ptr[3])
              else
              begin
                Str := copy(PChar(@Ptr[5]), 1, Ptr[3] - 1);
                if Ptr[4] = $01 then // Cyrillic
                  iso2win(Str);
                Providers[RecordCount] := Str;
              end;
            end;
            Ptr1 := @Ptr[4 + Ptr[3]];
            if Ptr1[0] > 0 then
            begin
              if Ptr1[1] >= $20 then
                Services[RecordCount] := copy(PChar(@Ptr1[1]), 1, Ptr1[0])
              else
              begin
                Str := copy(PChar(@Ptr1[2]), 1, Ptr1[0] - 1);
                if Ptr1[1] = $01 then // Cyrillic
                  iso2win(Str);
                Services[RecordCount] := Str;
              end;
            end;
          end;
        $4C: //Time-shifted service
          begin
            RecordCount := RecordCount + 1;
            SIDs[RecordCount] := SIDID;
            CAmode[RecordCount] := CAm;
            Providers[RecordCount] := '';
            Services[RecordCount] := 'Time-shifted ' + IntToStr(Ptr[2] * $100 + Ptr[3]);
          end;
      end;
      DescrLen := DescrLen + Ptr[1] + DESCR_GEN_LEN;
      Ptr := @Ptr[Ptr[1] + DESCR_GEN_LEN];
    end;
  end;
end;

function NowPlusMillis(Millis: Longint): TDateTime;
begin
  Result := Now + EncodeTime(0, 0, Millis div 1000, Millis mod 1000);
end;

{*******************************************************************}

function DVBGetTransporderInfo(TPInfo: PTransporderInfo): Integer;
  function FindPMT(SID:Integer): TPMTSection;
  var
    I: Integer;
    Section: TPMTSection;
  begin
    Result := nil;
    for I := 0 to PMTSectionList.Count - 1 do
    begin
      Section := PMTSectionList[I];
      if Section.ScannedSID = SID then
      begin
        Result := Section;
        Exit;
      end;
    end;
  end;
  procedure FreeSections;
  var
    I: Integer;
  begin
    if PATSection <> nil then
    begin
      PATSection.Free;
      PATSection := nil;
    end;
    if SDTSection <> nil then
    begin
      SDTSection.Free;
      SDTSection := nil;
    end;
    for I := 0 to PMTSectionList.Count - 1 do
    begin
      PMTSection := PMTSectionList[I];
      PMTSection.Free;
    end;
    PMTSectionList.Free;
  end;

var
  I, K: Integer;
  ChannelCount: Integer;
  PMTPid: Integer;
  Section: TPMTSection;
begin
//AddLog('Start Scan....');
  PMTSectionList := TList.Create;
  try
    ScanStop := False;
    TPInfo.ChannelCount := 0;
    ChannelCount := 0;
    Result := 1;
    PATSection := TPATSection.Create;
    PATSection.StopTime := NowPlusMillis(5000);
    PATSection.InitSection;
    while PATSection.Start(PID_PAT) = False do
    begin
      if ScanStop or (Now > PATSection.StopTime) then Exit;
      Sleep(50);
      Application.ProcessMessages;
    end;
//AddLog('Start PAT Section');
    PATSection.StopTime := NowPlusMillis(PAT_WAIT_TIME);
    while (Now < PATSection.StopTime) and (PATSection.Active) do
    begin
      if ScanStop then Exit;
      Sleep(30);
      Application.ProcessMessages;
    end;
//AddLog('Stop PAT Section ' + BoolToStr(PATSection.Active));
    if PATSection.Active then
    begin
      PATSection.StopScan;
      if PATSection.RecordCount = 0 then Exit;
    end;
//AddLog('Stop PAT Section Record Count = ' + IntToStr(PATSection.RecordCount));
    TPInfo.TSID := PATSection.TSID;
//    TransporderInfo.NetID := PATSection.NetworkPID;
    SDTSection := TSDTSection.Create;
    SDTSection.StopTime := NowPlusMillis(5000);
    SDTSection.InitSection;
    SDTSection.SearchedTSID := PATSection.TSID;
    while SDTSection.Start(PID_SDT) = False do
    begin
      if ScanStop or (Now > SDTSection.StopTime) then Exit;
      Sleep(50);
      Application.ProcessMessages;
    end;
//AddLog('Start SDT Section');
    SDTSection.StopTime := NowPlusMillis(SDT_WAIT_TIME);
    for I := 1 to PATSection.RecordCount do
    begin
      PMTSection := TPMTSection.Create;
      PMTSection.StopTime := NowPlusMillis(SDT_WAIT_TIME);
      PMTSection.InitSection;
      PMTSection.ScannedSID := PATSection.SIDs[I];
      PMTPid := PATSection.PMTs[I];
      while PMTSection.Start(PMTPid) = False do
      begin
        if ScanStop or (Now > PMTSection.StopTime) then Exit;
        for K := PMTSectionList.Count - 1 downto 0 do
        begin
          Section := PMTSectionList[K];
          if Section.Active and (Now > Section.StopTime) then
//begin
            Section.StopScan;
//AddLog('Error PMT Section PID = ' + IntToStr(Section.ScannedPID));
//end;
        end;
        Sleep(50);
        Application.ProcessMessages;
      end;
//AddLog('Start PMT Section PID = ' + IntToStr(PMTPid));
      PMTSection.StopTime := NowPlusMillis(PMT_WAIT_TIME);
      PMTSectionList.Add(PMTSection);
    end;
    while (Now < SDTSection.StopTime) and (SDTSection.Active) do
    begin
      if ScanStop then Exit;
      Sleep(30);
      Application.ProcessMessages;
    end;
//AddLog('Stop SDT Section ' + BoolToStr(SDTSection.Active));
    if SDTSection.Active then
    begin
      SDTSection.StopScan;
      if SDTSection.RecordCount = 0 then Exit;
    end;
    for K := PMTSectionList.Count - 1 downto 0 do
    begin
      Section := PMTSectionList[K];
      while (Now < Section.StopTime) and (Section.Active) do
      begin
        if ScanStop then Exit;
        Sleep(30);
        Application.ProcessMessages;
      end;
//AddLog('Stop PMT Section PID = ' + IntToStr(Section.ScannedPID));
      if Section.Active then Section.StopScan;
    end;
//AddLog('Stop SDT Section Record Count = ' + IntToStr(SDTSection.RecordCount));
    TPInfo.wNetID := SDTSection.NetworkID;
    for I := 1 to SDTSection.RecordCount do
    begin
      PMTSection := FindPMT(SDTSection.SIDs[I]);
      if PMTSection <> nil then
      begin
//AddLog('PMT Section PID = ' + IntToStr(PMTSection.ScannedPID));
//AddLog('PMT Section Name = ' + SDTSection.Services[I]);
//AddLog('PMT Section PCRPid = ' + IntToStr(PMTSection.PCRPid));
//AddLog('PMT Section VideoPID = ' + IntToStr(PMTSection.VideoPID));
//AddLog('PMT Section AudioPIDCount = ' + IntToStr(PMTSection.AudioPIDCount));
//for K := 1 to PMTSection.AudioPIDCount do
//AddLog('PMT Section AudioPID[' + IntToStr(K) + '] = ' + IntToStr(PMTSection.AudioPIDs[K]));
        if (PMTSection.PCRPid <> 0) or (PMTSection.VideoPID <> 0) or (PMTSection.AudioPIDCount > 0) then
        begin
          TPInfo.Channel[ChannelCount].wSID := SDTSection.SIDs[I];
          TPInfo.Channel[ChannelCount].wPMT := PMTSection.ScannedPID;
          if PMTSection.VideoPID <> 0 then
            TPInfo.Channel[ChannelCount].wType := 1
          else if PMTSection.AudioPIDCount > 0 then
            TPInfo.Channel[ChannelCount].wType := 2
          else
            TPInfo.Channel[ChannelCount].wType := 0;
          TPInfo.Channel[ChannelCount].wScrambled := SDTSection.CAmode[I];
          for K := 1 to Length(SDTSection.Services[I]) do
            TPInfo.Channel[ChannelCount].ChannelName[K - 1] := SDTSection.Services[I,K];
          TPInfo.Channel[ChannelCount].ChannelName[Length(SDTSection.Services[I])] := #0;
          for K := 1 to Length(SDTSection.Providers[I]) do
            TPInfo.Channel[ChannelCount].ProviderName[K - 1] := SDTSection.Providers[I,K];
          TPInfo.Channel[ChannelCount].ProviderName[Length(SDTSection.Providers[I])] := #0;
          TPInfo.Channel[ChannelCount].wPCR := PMTSection.PCRPid;
          TPInfo.Channel[ChannelCount].wVideo := PMTSection.VideoPID;
          for K := 1 to PMTSection.AudioPIDCount do
          begin
            TPInfo.Channel[ChannelCount].Audio.wPID[K - 1] := PMTSection.AudioPIDs[K];
            TPInfo.Channel[ChannelCount].Audio.Language[K - 1,0] := PMTSection.AudioLanguages[K,1];
            TPInfo.Channel[ChannelCount].Audio.Language[K - 1,1] := PMTSection.AudioLanguages[K,2];
            TPInfo.Channel[ChannelCount].Audio.Language[K - 1,2] := PMTSection.AudioLanguages[K,3];
          end;
          TPInfo.Channel[ChannelCount].Audio.wCount := PMTSection.AudioPIDCount;
          TPInfo.Channel[ChannelCount].wAC3PID := PMTSection.AC3Pid;
          TPInfo.Channel[ChannelCount].wTeletext := PMTSection.TeletextPID;
          for K := 1 to PMTSection.CACount do
          begin
            TPInfo.Channel[ChannelCount].ECM.wPID[K - 1] := PMTSection.CAECMs[K];
            TPInfo.Channel[ChannelCount].ECM.wCAType[K - 1] := PMTSection.CAIDs[K];
          end;
          TPInfo.Channel[ChannelCount].ECM.wCount := PMTSection.CACount;
          ChannelCount := ChannelCount + 1;
        end;
      end;
    end;
    TPInfo.ChannelCount := ChannelCount;
  finally
    FreeSections;
  end;
//AddLog('Stop Scan....');
end;

initialization
  InitializeCriticalSection(LockDVBSectionList);
  DVBSectionList := TList.Create;
  ZeroMemory(@TransporderInfo, SizeOf(TransporderInfo));
finalization
  DVBSectionList.Free;

end.
